.IFNDEF CALIBRATION
.EQU CALIBRATION=1

;#############################################################################
;#                                                                           #
;#         Interrupts must not enabled when running calibration!             #
;#                                                                           #
;#############################################################################      

.INCLUDE "unifiedio.inc"
.INCLUDE "calibration_cfg.inc"

.EQU CAL_NOMINAL = FCLOCK / CAL_REFCLK

.IF CAL_NOMINAL < 10
	.WARNING "Calibration reference clock is to high in relation to system clock"
.ENDIF

.IF CAL_NOMINAL > 32767
	.WARNING "Calibration reference clock is to low in relation to system clock"
.ENDIF

.IF CAL_NOMINAL * CAL_REFCLK != FCLOCK
	.WARNING "System clock is not an integer multiple of calibration reference clock"
.ENDIF 

.EQU CAL_TIMEOUT = Int(1.5 * CAL_NOMINAL) 
.EQU CAL_CORRECT = CAL_TIMEOUT - CAL_NOMINAL

.EQU CAL_BOTTOM_BIT = 0
.EQU CAL_BOTTOM = 1<<CAL_BOTTOM_BIT 
.EQU CAL_TOP_BIT = 1
.EQU CAL_TOP = 1<<CAL_TOP_BIT
.EQU CAL_SKIPUP_BIT = 2
.EQU CAL_SKIPUP = 1<<CAL_SKIPUP_BIT
.EQU CAL_SKIPDOWN_BIT = 3
.EQU CAL_SKIPDOWN = 1<<CAL_SKIPDOWN_BIT

.MACRO CAL_MEASURE

waitforhigh:	
	sbiw R25:R24,1						; 2  2
	sbis CAL_INPUTREG,CAL_INPUTBIT		; 2  1
	brpl waitforhigh					; 	 2
										;   ..
										;	 5											
waitforlow:                            
	sbiw R25:R24,1						; 2  2
	sbic CAL_INPUTREG,CAL_INPUTBIT		; 2  1
	brpl waitforlow						;    2
										;   ..
										;    5
		                              	;-- -- 
.ENDMACRO								; 8 10 

.CSEG

;<-Zeroflag set on error, cleared on success
;  On success OSCCAL will contain found calibration value,
;  initial OSCCAL value otherwise.
cal_calibrate:
	push R16
	push R17
	push R18
	push R19
	push R20
	push R21
	push R22
	push R23
	push R24
	push R25

	ldi R16,0		  				;init state machine
	UIN R17,OSCCAL					;get initial value
	mov R18,R17						;last is initial
	mov R19,R17						;best is initial 
	ldi R20,Low($7fff)				;default best result to worst
	ldi R21,High($7fff)				;possible
	ldi R22,Low(CAL_TIMEOUT)		;timeout value for five
	ldi R23,High(CAL_TIMEOUT)		;periods of refence clock

	_cal_loop:
	ldi R24,Low(CAL_TIMEOUT / 5)	;lock to refclock in one period
	ldi R25,High(CAL_TIMEOUT / 5)
	CAL_MEASURE
	brmi _cal_error					;check timeout condition
	movw R25:R24,R23:R22			;measure over five periods, check
	CAL_MEASURE						;timeout condition after each
	brmi _cal_error
	nop
	CAL_MEASURE
	brmi _cal_error
	nop
	CAL_MEASURE
	brmi _cal_error
	nop
	CAL_MEASURE
	brmi _cal_error
	nop
	CAL_MEASURE
	brmi _cal_error
	
	subi R24,Low(CAL_CORRECT * 2)	;check upper limit of calibration window
	sbci R25,High(CAL_CORRECT * 2)
	brcc _cal_error					;outside

	subi R24,Low(-CAL_CORRECT)		;finally get the deviation from nominal
	sbci R25,High(-CAL_CORRECT)
	brmi _cal_tofast				;we are to fast
	brne _cal_toslow				;we are to slow
	rjmp _cal_success				;we found best possible solution

	_cal_error:
	ldi R19,1						;restore initial value in gentle slope
	cp R17,R18
	brcc _cal_slope
	neg R19
	_cal_slope:
	cp R17,R18
	breq _cal_done
	add R18,R19
	UOUT OSCCAL,R18
	rjmp _cal_slope

	_cal_tofast:
	sbr R16,CAL_TOP					;manage state machine flags
	cbr R16,CAL_SKIPDOWN
	sbrc R16,CAL_SKIPUP_BIT
	cbr R16,CAL_SKIPUP|CAL_BOTTOM

	com R24							;abs()
	com R25
	adiw R25:R24,1

	cp R24,R20						;compare to best previous result
	cpc R25,R21
	brcc _cal_bad_up				;it was better

	movw R21:R20,R25:R24			;else record new best 
	mov R19,R18

	_cal_bad_up:
	sbrc R16,CAL_BOTTOM_BIT			;check opposite direction
	rjmp _cal_usebest				;already scanned

	subi R18,1						;scan downward
	brcs _cal_error					;but not below zero

.IFDEF CAL7							;for 8bit OSCCAL register

	cpi R18,127						;expect discontinuity at this point
	brne _cal_normaldown			;if not there

	subi R18,32						;else big value step, but small for clock
	sbr R16,CAL_SKIPDOWN			;record the jump

	_cal_normaldown:

.ENDIF

	UOUT OSCCAL,R18
	rjmp _cal_loop

	_cal_toslow:
	sbr R16,CAL_BOTTOM				;manage state machine flags
	cbr R16,CAL_SKIPUP
	sbrc R16,CAL_SKIPDOWN_BIT
	cbr R16,CAL_SKIPDOWN|CAL_TOP

	cp R24,R20						;compare to best previous result
	cpc R25,R21
	brcc _cal_bad_down				;it was better

	movw R21:R20,R25:R24			;else record new best 
	mov R19,R18

	_cal_bad_down:
	sbrc R16,CAL_TOP_BIT			;check opposite direction
	rjmp _cal_usebest				;already scanned

	subi R18,-1						;scan upward

.IFDEF CAL7							;for 8bit OSCCAL register

	brcc _cal_error					;dont scan above 255

	cpi R18,96						;expect discontinuity 32 steps beyond this
	brne _cal_normalup				;if not there

	subi R18,-32					;else big value step, but small for clock
	sbr R16,CAL_SKIPUP				;record the jump

	_cal_normalup:

.ELSE								;for 7bit OSCCAL regeister

	brmi _cal_error					;dont scan above 127

.ENDIF

	UOUT OSCCAL,R18
	rjmp _cal_loop

	_cal_usebest:
	UOUT OSCCAL,R19					;set OSCCAL value for best result

	_cal_success:
	clz

	_cal_done:
	pop R25
	pop R24
	pop R23
	pop R22
	pop R21
	pop R20
	pop R19
	pop R18
	pop R17
	pop R16
	ret

.IFDEF CAL_EEPADDR

;<-R16: OSCCAL value, stored if available, else default
cal_load:
	push R17

	ldi R16,Low(CAL_EEPADDR)		;get OSCCAL value
	out EEARL,R16
	ldi R16,High(CAL_EEPADDR)
	out EEARH,R16
	sbi EECR,EERE
	in R17,EEDR

	ldi R16,Low(CAL_EEPADDR+1) 		;and it's one's complement
	out EEARL,R16
	ldi R16,High(CAL_EEPADDR+1)
	out EEARH,R16
	sbi EECR,EERE
	in R16,EEDR

	com R16
	cp R16,R17						;valid?
	breq _cl_done					;yes

	ldi R16,CAL_DEFOSCCAL			;else use default				

	_cl_done:
	pop R17
	ret

;->R16: OSCCAL value to store
cal_store:
	push R16
	push R17
	
	mov R17,R16
	 
	ldi R16,Low(CAL_EEPADDR) 		;store OSCCAL value
	out EEARL,R16
	ldi R16,High(CAL_EEPADDR)
	out EEARH,R16
	out EEDR,R17
	sbi EECR,EEMPE
	sbi EECR,EEPE

	_cs_waiteep0:
	sbic EECR,EEPE					;wait for write completion
	rjmp _cs_waiteep0

	com R17							;store the one's complement
	ldi R16,Low(CAL_EEPADDR+1) 		;to mark stored value as valid
	out EEARL,R16
	ldi R16,High(CAL_EEPADDR+1)
	out EEARH,R16
	out EEDR,R17
	sbi EECR,EEMPE
	sbi EECR,EEPE

	_cs_waiteep1:					;wait for write completion
	sbic EECR,EEPE
	rjmp _cs_waiteep1

	pop R17
	pop R16
	ret

.ENDIF

.ENDIF
